
import java.net.*;
import java.io.*;
import java.util.*;

import java.util.regex.Pattern;
import java.util.regex.Matcher;

class Queue extends Vector
{
	public Queue(int capacity)
	{
		super(capacity);
	}

	public synchronized boolean enqueue(Object o)
	{
		if(!contains(o))
		{
			addElement(o);
			return true;
		}
		return false;
	}

	public synchronized Object dequeue()
	{
		return remove(0);
	}
};

class SpiderParser
{

	public String url;
	public URL url_obj;
	public Vector links;
	public long doc;
	public long dom;
	public int size;

	public String regex;
	public Pattern reg;

	public SpiderParser()
	{
		links = new Vector(10);

		regex = "(h|H)(r|R)(e|E)(f|F)=[a-zA-Z0-9:\\-\\/\\.\\?\\=\\&\"']*";
		reg = Pattern.compile(regex);
	}

	public boolean parse(URL l)
	{
		links.clear();

		url_obj = l;

		url = url_obj.toString();

		try
		{
			URLConnection conn;
			try
			{
				 conn = url_obj.openConnection();
			}
			catch (OutOfMemoryError e)
			{
				XMLSpider.reportOutOfMem("parser: open conn ("+url+")");
				return false;

			}

			conn.setAllowUserInteraction(false);

			doc = conn.getDate();
			dom = conn.getLastModified();

			String content = getPage();

			// parse content
			size = content.length();

// 			String findAnchors = "<a[^>]*>";
// 			String findHref = "href=([a-zA-Z\\./\"%\\?])+";
// 			Pattern anchors = Pattern.compile(findAnchors);
// 			Pattern hrefs = Pattern.compile(findHref);

			Matcher matcher;
			try
			{
				matcher = reg.matcher(content);
			}
			catch (OutOfMemoryError e)
			{
				XMLSpider.reportOutOfMem("parser: making matcher ("+url+")");
				return false;

			}


			String dir = url_obj.toString();
			dir = dir.substring(0,dir.lastIndexOf('/')+1);

			// find and print all matches
			try
			{
				while (matcher.find())
				{
					String s = matcher.group();

					s = reform(s,dir);

					if ( s!=null && !links.contains(s) )
					{
						links.add(s);
						//System.out.println("link: "+s);
					}
				}
			}
			catch (OutOfMemoryError e)
			{
				XMLSpider.reportOutOfMem("parser: matching ("+url+")");
				return false;

			}

		}
		catch (Exception e)
		{
			//System.err.println("ERROR: couldn't open url: "+url);
			return (false);
		}
		catch (OutOfMemoryError e)
		{
			XMLSpider.reportOutOfMem("parser ("+url+")");
			return false;

		}

		return (true);
	}

	public String getPage()
	{
		String content = new String("");
		int numRead = 0;
		try
		{
			URLConnection conn = url_obj.openConnection();
			conn.setAllowUserInteraction(false);

			InputStream urlStream = url_obj.openStream();
			String type = conn.guessContentTypeFromStream(urlStream);

// 			if (type == null)
// 			{
// 				System.err.println("ERROR: url has no known type: "+url);
// 				return false;
// 			}
// 			if (type.compareTo("text/html") != 0)
// 			{
// 				System.err.println("ERROR: url has no text/html type: "+url);
// 				return false;
// 			}

			// search the input stream for links
			// first, read in the entire URL
			byte b[] = new byte[1000];
			numRead = urlStream.read(b);
			content = new String(b, 0, numRead);

			while (numRead != -1)
			{
				numRead = urlStream.read(b);

				if (numRead != -1)
				{
					String newContent = new String(b, 0, numRead);
					content += newContent;
				}
			}
			urlStream.close();
		}
		catch (Exception e)
		{
			return null;
		}
		catch (OutOfMemoryError e)
		{
			XMLSpider.reportOutOfMem("getting page ("+url+") size: "+numRead+" content length: "+content.length());
			return null;
		}

		return content;

	}

	// check and nice-ify url string
	public String reform(String s, String path)
	{
		s = s.trim();

		s = s.substring(5);

		if(s.startsWith("\""))
		{
			s = s.substring(1);
		}
		if(s.endsWith("\""))
		{
			s = s.substring(0,s.length()-1);
		}

		if( -1 != s.indexOf("\"") )
			s = s.substring(0,s.indexOf("\""));

		if( s.startsWith("https") || s.startsWith("ftp") || s.startsWith("mailto") )
			return null;

		// some legal endings
		if( !( s.endsWith(".html") || s.endsWith(".htm") || s.endsWith(".php") || s.endsWith("/") ) )
			return null;

		if( s.startsWith("./") ) // add the current path in front
			s = path+s.substring(2);

		if( s.startsWith("../") )
		{
			do
			{
				// cut off the end dir name
				path = path.substring(0,path.lastIndexOf('/',path.length()-2)+1);
				// cut off front 3 chars "../"
				s = s.substring(3);
			} while( s.startsWith("../") );
			s = path+s;
		}

		if( s.startsWith("/") )
		{
			path = path.substring(0,path.indexOf('/',7)); // at least "http://"
			s = path+s;
		}


		if( !s.startsWith("http://") ) // a relative path!
			s = path+s;

		try
		{
			URL m = new URL(s);
		}
		catch (MalformedURLException e)
		{
			System.err.println(">> Malformed url: "+s+" <<");
			return null;
		}

		return (s);
	}

};

class XMLWriter
{
	public PrintWriter out;
	private int count;
	private int counter;
	private int numtasks;

	public synchronized int getCount()
	{
		return counter;
	}

	public synchronized void incCount()
	{
		counter++;
	}

	public synchronized void incTasks()
	{
		numtasks++;
	}

	public synchronized int getTasks()
	{
		return numtasks;
	}

	public XMLWriter(String filename)
	{
		count = 0;
		counter = 0;
		try
		{
			//out = new PrintWriter(new BufferedWriter(new FileWriter(filename)));
			out = new PrintWriter(new FileWriter(filename));
		}
		catch (IOException e)
		{
			System.err.println("Couldn't open output file: "+filename);
			System.exit(1);
		}
	}


	public void start()
	{
		out.println("<ggle>");
		out.flush();
	}

	public synchronized void write(SpiderParser p)
	{
		count++;
		if(count%1000 == 0)
		{
			System.out.print(count);
			System.out.println(": "+p.url);
		}
		else if(count%100 == 0)
		{
			System.out.print(count);
			System.out.flush();
		}
		else if(count%20 == 0)
		{
			System.out.print('.');
			System.out.flush();
		}

		out.print("<entry");
		out.print(" url=\""); out.print(xmlescape(p.url)); out.print("\"");
		out.print(" doc=\""); out.print(p.doc); out.print("\"");
		out.print(" dom=\""); out.print(p.dom); out.print("\"");
		out.print(" size=\""); out.print(p.size); out.print("\"");
		out.println(">");

		for( int i=0; i<p.links.size(); i++)
		{
			out.println("        <link url=\""+xmlescape((String)p.links.elementAt(i))+"\"/>");
		}

		out.println("</entry>");

		if( count%50 == 0) // only flush after 50 entries
			out.flush();
	}

	public String xmlescape(String s)
	{
		s = s.replaceAll("&","&amp;");
		s = s.replaceAll("'","&apos;");
		s = s.replaceAll("\"","&quot;");
		s = s.replaceAll(">","&gt;");
		s = s.replaceAll("<","&lt;");
		return s;
	}

	public void finish()
	{
		out.println("</ggle>");
		out.flush();
	}
};

class XMLSpiderWorker extends Thread
{
	public Queue vertices;
	public Set visited;
	public long delay;
	public int maxVertices;
	public int hits;

	public XMLWriter out;

	private URL starturl;

	public static final String DISALLOW = "Disallow:";


	public XMLSpiderWorker(XMLWriter o, Set s, long d, int max)
	{
		out = o;
		maxVertices = max;
		delay = d;
		hits = 0;

		visited = s;
		vertices = new Queue(10);

		System.out.print("--making worker: delay=");
		System.out.print(d);
		System.out.print(" maxVertices=");
		System.out.print(max);
		System.out.println(" --");
	}

	public void process(URL start)
	{
		out.incTasks(); //count myself as a task
		System.out.println("--starting thread for url: "+start.toString()+" --");
		starturl = start;
		this.start();
	}

	public void run()
	{
		//System.out.println("-- running! --");

		URL v;
		String s;
		SpiderParser p = new SpiderParser();

		//vertices.enqueue(starturl);
		vertices.enqueue(starturl.toString());

		try
		{
			while( !vertices.isEmpty() && out.getCount() < maxVertices )
			{
				//v = (URL)vertices.dequeue();
				s = (String)vertices.dequeue();
				try
				{
					v = new URL(s);
				}
				catch(MalformedURLException e)
				{
					continue;
				}

				if( isOKtoVisit( v, s ) )
				{
					try
					{
						if(!p.parse(v))
						{
							//System.err.println("ERROR: could not parse url: "+v.toString());
							continue;
						}
					} catch(OutOfMemoryError e)
					{
						System.err.print("run: parsing ("+starturl.toString()+")");
					}

					System.gc();

					hits++; //internal counter
					out.incCount();
					visit(v,p);
					try
					{
						synchronized(visited)
						{
							visited.add(s);
						}
					} catch(OutOfMemoryError e)
					{
						XMLSpider.reportOutOfMem("run: visited ("+starturl.toString()+")");
					}

					System.gc();

					//put only 70% of outgoing links into queue
					int j=0;
					for( int i=0; (j < (int)(p.links.size()*.7)) && (i < p.links.size() ); i++ )
					{
						try
						{
							//URL n = new URL((String)p.links.elementAt(i));
							//vertices.enqueue(n);

							// only add new links to queue when we still have to gather new links
							if(     ((vertices.size()+hits)<=(maxVertices/out.getTasks())) &&
								(vertices.size()<=(maxVertices/out.getTasks()/10)) &&
								vertices.enqueue((String)p.links.elementAt(i)) )
									j++; //only count correctly added vertices
						}
						//catch (MalformedURLException e) {}
						catch (OutOfMemoryError e)
						{
							XMLSpider.reportOutOfMem("run: new url ("+starturl.toString()+")");
						}
					}
					v=null;
					s=null;

					System.gc();

					try
					{
						sleep(delay);
					}
					catch (InterruptedException e) {}
				}
				else
				{
					//System.out.println("-- not allowed for url: "+v.toString()+" --");
				}

				System.gc();
			}
		}
		catch(OutOfMemoryError e)
		{
			XMLSpider.reportOutOfMem("run ("+starturl.toString()+")");
		}

		System.out.println("-- done! ("+starturl.toString()+") --");
	}

	private void visit(URL v, SpiderParser p)
	{
		// print out the XML entry about this site
		out.write(p);
	}

	private boolean isOKtoVisit(URL link, String s) {
		// Return false if it's not HTTP protocol
		if (!link.getProtocol().equals("http"))
		{
			//System.err.println("WARN: url not of right proto: "+link.toString());
			return(false);
		}
		synchronized(visited)
		{
			if (visited.contains(s))
			{
				//System.err.println("WARN: url already visited: "+link.toString());
				return(false);
			}
		}
		if (!robotSafe(link))
		{
			//System.err.println("WARN: url not robotable: "+link.toString());
			return(false);
		}
		else return(true);
	}

	boolean robotSafe(URL url)
	{
		try
		{
			String strHost = url.getHost();

			// form URL of the robots.txt file
			String strRobot = "http://" + strHost + "/robots.txt";
			URL urlRobot;
			try
			{
				urlRobot = new URL(strRobot);
			}
			catch (MalformedURLException e)
			{
				// something weird is happening, so don't trust it
				return false;
			}

			String strCommands = "";
			int numRead = 0;
			try
			{
				InputStream urlRobotStream = urlRobot.openStream();

				// read in entire file
				byte b[] = new byte[1000];
				numRead = urlRobotStream.read(b);

				if( numRead == -1 ) // no data?
					return true;

				strCommands = new String(b, 0, numRead);
				while (numRead != -1)
				{
					numRead = urlRobotStream.read(b);
					if (numRead != -1)
					{
						String newCommands = new String(b, 0, numRead);
						strCommands += newCommands;
					}
				}
				urlRobotStream.close();
			}
			catch (IOException e)
			{
				// if there is no robots.txt file, it is OK to search
				return true;
			}
			catch (OutOfMemoryError e)
			{
				String s = "reading robots.txt ("+urlRobot.toString()+"), strlength: "+strCommands.length();
				XMLSpider.reportOutOfMem(s);
				return false;
			}

			// assume that this robots.txt refers to us and
			// search for "Disallow:" commands.
			String strURL = url.getFile();
			int index = 0;
			while ((index = strCommands.indexOf(DISALLOW, index)) != -1)
			{
				index += DISALLOW.length();
				String strPath = strCommands.substring(index);
				StringTokenizer st = new StringTokenizer(strPath);

				if (!st.hasMoreTokens())
					break;

				String strBadPath = st.nextToken();

				// if the URL starts with a disallowed path, it is not safe
				if (strURL.indexOf(strBadPath) == 0)
					return false;
			}

			return true;
		}
		catch (OutOfMemoryError e)
		{
			XMLSpider.reportOutOfMem("robotsafe ("+url.toString()+")");
			return false;
		}
	}
};

public class XMLSpider
{
	public static Vector workers;

	public static synchronized void reportOutOfMem(String s)
	{
		System.err.print('!');System.err.flush();
// 		System.err.print("-- oom: ");
// 		System.err.print(s);
// 		System.err.println(" :: queues: ");
//
// 		int w;
//
// 		for(int i=0; i<workers.size(); i++)
// 		{
// 			w = ((XMLSpiderWorker)workers.elementAt(i)).visited.size();
// 			System.err.print(i);
// 			System.err.print(" size ");
// 			System.err.println(w);
// 		}
// 		System.err.println(" --");
// 		System.err.flush();
	}

	public static void main(String argv[])
	{
		if (argv.length != 1)
		{
			System.err.println("Usage: cmd max-num");
			System.exit(1);
		}

		int NumSites = (new Integer(argv[0])).intValue();
		long sleepytime = 2500; //millisecs

		Vector startsites = new Vector(10);
		try
		{
			startsites.add(new URL("http://homepages.cwi.nl/~mk/onderwijs/adt/index.html"));
			startsites.add(new URL("http://www.mysql.com/"));
			startsites.add(new URL("http://www.postgresql.com/"));
			startsites.add(new URL("http://openil.sourceforge.net/index.php"));
			startsites.add(new URL("http://citeseer.ist.psu.edu/253431.html"));
			startsites.add(new URL("http://eecs.iu-bremen.de/index.html"));
			startsites.add(new URL("http://carol.wins.uva.nl/~jverbeek/mlph/index_en.html"));
			startsites.add(new URL("http://www.science.uva.nl/research/cs.html/index.html"));
			startsites.add(new URL("http://www.syseng.anu.edu.au/rsl/sub/2-Systems/Submersibles.html"));
			startsites.add(new URL("http://www.science.uva.nl/research/scs/edu/scivis/index.html"));
			startsites.add(new URL("http://www.python.org/index.html"));
			startsites.add(new URL("http://doc.trolltech.com/index.html"));
			startsites.add(new URL("http://irrlicht.sourceforge.net/"));
			startsites.add(new URL("http://www.libsdl.org/index.php"));
			startsites.add(new URL("http://www.php.net/"));
			startsites.add(new URL("http://www.kernel.org/"));
			startsites.add(new URL("http://freshmeat.net/"));
			startsites.add(new URL("http://www.gnu.org/"));
			startsites.add(new URL("http://java.sun.com/j2se/1.4.2/docs/api/java/lang/System.html"));
			startsites.add(new URL("http://members.net-tech.com.au/alaneb/opengl_links.html"));
		}
		catch (MalformedURLException e)
		{
			System.err.println("Something went wrong making the start urls!");
			System.exit(1);
		}

		// make shared resources

		Set s = Collections.synchronizedSet(new HashSet(4*NumSites));

		XMLWriter out = new XMLWriter("output.xml");
		out.start();

		// end shared resources

		workers = new Vector(startsites.size());

		for(int i=0; i<startsites.size(); i++)
			workers.add( new XMLSpiderWorker(out,s,sleepytime,NumSites) );

		System.out.print("Made ");
		System.out.print(startsites.size());
		System.out.println(" threads, starting...");

		for(int i=0; i<startsites.size(); i++)
			((XMLSpiderWorker)workers.elementAt(i)).process((URL)startsites.elementAt(i));

		System.out.print("Started ");
		System.out.print(startsites.size());
		System.out.println(" threads, joining...");

		try
		{
			for(int i=0; i<workers.size(); i++)
				((XMLSpiderWorker)workers.elementAt(i)).join();
		}
		catch (InterruptedException e)
		{
			System.err.println("The join calls got interrupted! exiting...");
			System.exit(1);
		}

		System.out.println("Threads finished!");

		// all threads finished
		out.finish();
	}

};
