import java.util.*;
import java.io.*;
import java.sql.*;




/*-------------------------------------------------------------- LoadGGLE class
|																			  |
|				Class to load the GGLE db from a xml file					  |
|																			  |
|																			  |
|																			  |
| 													by Felix Hageloh 01/10/04 |
\----------------------------------------------------------------------------*/
public class LoadGGLE
{

	public static final String FILE = "output.xml";

	// CHANGE DRIVER STRING HERE
	public static final String DRIVER =
		"nl.cwi.monetdb.jdbc.MonetDriver";
		//"com.mysql.jdbc.Driver";

	// CHANGE DB CONNECTION STRING HERE
	public static final String DB =
		"jdbc:monetdb://localhost/ggle";
		//"jdbc:mysql://localhost/ggle?user=ggle&password=gsa#1";


	/*------------------------------------------------------------------- main()
	|																		   |
	| main function															   |
	\-------------------------------------------------------------------------*/
	public static void main (String argv[])
	{

		// --------------------------------------------- Open the db connection

		Connection con; 		// tho connection pointer
		DatabaseMetaData d; String dbName; String user;

		try
		{
			Class.forName(DRIVER);
			con = DriverManager.getConnection(DB, "ggle", "gsa#1");

			//con.setAutoCommit( false );
			d = con.getMetaData();
			dbName =
				d.getDatabaseProductName() + " " +  d.getDatabaseProductVersion();
			user = d.getUserName();
		}
		// --- connection could not be opened -> exit
		catch (Exception e)
		{
			System.out.println(e);
			return;
		}

		System.out.println(
			"Connected to " + dbName + " as \"" + user
			 + "\". Auto commit off");

		// -------------------------- create a new xml parser on the given file
		XMLparser xml = new XMLparser(FILE);

		// ---- file couldn't be opned -> exit
		if( xml == null )
			return;

		int stat = 0;		// status report from nextEntry()
		int EntryId = 0; 	// uid for Entry table
		int LinkId = 1;		//uid for Link Table

		// --------------------------------------- get the first entry
		try
		{
			stat = xml.nextEntry();	// get the next entry
			EntryId++;
		}
		catch( IOException e )
		{
			System.out.println(
			"ERROR READING FROM FILE!!\n IO EXCEPTION: " + e);
			stat = -1;
		}


		// --- variable declarations
		String url = null;
		String domain = null;
		String path = null;
		String page	= null;
		String[] parts = new String[2];
		PreparedStatement pstmt;
		Statement stmt;
		ResultSet linksTo, value;
		SQLWarning warn;
		long start, end = 0;


		System.out.println("Loading data ... ");

		//clearDB();

		// -------------------------------------- loop through all entries
		start = System.currentTimeMillis();
		while( stat != -1 )
		{

			//System.out.println("stat = " + stat);

			// --- error reading next entry -> skip
			if ( stat == 0 )
			{
				try
				{
					stat = xml.nextEntry();
				}
				catch(IOException e)
				{
					System.out.println(
						"ERROR READING FROM FILE!!\n IO EXCEPTION: " + e +
						"LAST ENTRY READ: " + url );
					stat = -1;
					break;
				}
				continue;
			}

			//  ------------------ split the url into domain path and page
			url = xml.getURL();

			int first = url.indexOf('/', 8);
			int last = url.lastIndexOf('/') + 1;

			if( first != -1 )
			{
				domain = url.substring( 0, first );
				path = url.substring( first, last );
				page = url.substring( last );
			}
			else
			{
				domain = url;
			}
			/*System.out.println("domain: " + domain);
			System.out.println("path: " + path);
			System.out.println("page: " + page);*/


			// ----------------------------------------- prepare the query
			try
			{
				pstmt = con.prepareStatement
				(
						"INSERT INTO Entry VALUES (?, ?, ?, ?, ?, ?, ?, ?)"

				);
				pstmt.setInt	(	1, EntryId			);
				pstmt.setString	(	2, domain  			);
				pstmt.setString	(	3, path				);
				pstmt.setString	(	4, page				);
				pstmt.setString	(	5, url				);
				pstmt.setDate	(	6, xml.getDOC()		);
				pstmt.setDate	(	7, xml.getDOM()		);
				pstmt.setInt	(	8, xml.getSIZE()	);

				// ------------------------------------- execute the query
				pstmt.executeUpdate();

				// ------------------------------------ check for warnings
				//warn = pstmt.getWarnings();

				//if( warn != null )
					//System.out.println(warn);
			}
			// --- error during insert -> skip
			catch( SQLException e )
			{
				System.out.println(
				"SQL ERROR: Skipping Entry " + url + "\nEXCEPTION: " + e);
				try
				{ stat = xml.nextEntry();} catch(Exception ex) { System.out.println("fail!"); return;};
				continue;
			}





			Vector L = xml.getLINKS();

			// ----------------------------------------- Loop through all links

			for (Enumeration links = L.elements() ; links.hasMoreElements();)
			{

				//  --------- split the link url into domain path and page
				url = (String) links.nextElement();


				first = url.indexOf('/', 8);
				last = url.lastIndexOf('/') + 1;

				if( first != -1 )
				{
					domain = url.substring( 0, first );
					path = url.substring( first, last );
					page = url.substring( last );
				}
				else
				{
					domain = url;
				}

				// ------------------------------------- prepare the query
				try
				{
					pstmt = con.prepareStatement
					(
						"INSERT INTO Link(LinkId, Entry, domain, path, page, url) "
						+ "  VALUES (?, ?, ?, ?, ?, ?)"

					);
					pstmt.setInt	(	1, LinkId 			);
					pstmt.setInt	(	2, EntryId 			);
					pstmt.setString	(	3, domain  			);
					pstmt.setString	(	4, path				);
					pstmt.setString	(	5, page				);
					pstmt.setString	(	6, url				);

					// -------------------------------- execute the query
					pstmt.executeUpdate();

					// ------------------------------- check for warnings
					//warn = pstmt.getWarnings();

					//if( warn != null )
						//System.out.println(warn);
				}
				// --- error during insert -> skip
				catch( SQLException e )
				{
					System.out.println(
					"SQL ERROR: Skipping Link " + url + "\nEXCEPTION: " + e);
					continue;
				}

				LinkId++; 				// increment the LinkId

			} // -- end link loop

			// ---------------------------------------- get the next entry
			try
			{
				stat = xml.nextEntry();	// get the next entry
				EntryId++;				// increment the EntryId
			}
			catch( IOException e )
			{
				System.out.println(
				"ERROR READING FROM FILE!!\n IO EXCEPTION: " + e +
				"LAST ENTRY READ: " + url );
				stat = -1;
			}

		} // --- end entry loop

		//--------------------------------------------------- update links
		try
		{
			stmt = con.createStatement();

			// ------------------------------------------ add foreign keys
			stmt.execute
			(
				"ALTER TABLE Link " +
				"	ADD FOREIGN KEY(Entry) REFERENCES Entry(EntryId);"
			);

			stmt.execute
			(
				"ALTER TABLE Link " +
				"	ADD FOREIGN KEY(LinksTo) REFERENCES Entry(EntryId);"
			);


			// -------------------------------- execute the query
			linksTo = stmt.executeQuery
			(
				"SELECT URL, LinksTo, LinkId FROM Link"
			);

			// ------------------------------- check for warnings
			//warn = stmt.getWarnings();

			//if( warn != null )
				//System.out.println(warn);

			while( linksTo.next() )
			{
				//System.out.println("\nlink: " + linksTo.getString(1));

				pstmt = con.prepareStatement
				(
					"SELECT EntryId " +
					"	FROM Entry " +
					"	WHERE URL = ?"
				);
				pstmt.setString	(	1, linksTo.getString(1) );

				value = pstmt.executeQuery();

				// ------------------ check if an entry was found
				if(!value.first() )
					continue;

				//warn = pstmt.getWarnings();

				//if( warn != null )
				//	System.out.println(warn);

				pstmt = con.prepareStatement
				(
					"UPDATE Link SET LinksTo = ? " +
					"	WHERE LinkId = ? "
				);
				pstmt.setInt(	1, value.getInt(1) );
				pstmt.setInt(	2, linksTo.getInt(3) );

				pstmt.executeUpdate();

			}
			end = System.currentTimeMillis();

			// ---------------------------------- clean up and close connection
			//con.commit();
			xml.cleanUp();
			con.close();
		}
		catch(Exception e)
		{
			System.out.println(e);
		}

		int min, sec, ms;

		Calendar cal = Calendar.getInstance();

		cal.setTime( new java.util.Date(end-start) );
		min = cal.get(Calendar.MINUTE);
		sec = cal.get(Calendar.SECOND);
		ms	= cal.get(Calendar.MILLISECOND);

		System.out.println(
			"Finished loading DB. Loaded " + (EntryId - 1) + " entries, " +
			(LinkId - 1) + " links in " + min + "min " + sec + "sec " + ms + "ms");
	}


	// ******************** stupid function that we shouldn't have to write!!!!!!

	public static void clearDB()
	{
		try
		{
			Class.forName("nl.cwi.monetdb.jdbc.MonetDriver");
			Connection con = DriverManager.getConnection(
				"jdbc:monetdb://localhost/ggle", "ggle", "gsa#1");


			Statement stmt = con.createStatement();

			stmt.execute("DROP TABLE Entry");
			stmt.execute("DROP TABLE Link");

			stmt.execute
			(
				"CREATE TABLE Entry "+
				"(" +
				"	EntryId		INT		UNIQUE 	NOT NULL, " +
				"	Domain		VARCHAR(50) 	NOT NULL, " +
				"	Path		VARCHAR(100) 	NOT NULL, " +
				"	Page		VARCHAR(50), " +
				"	URL		VARCHAR(200)	NOT NULL," +
				"	DOC		DATE, " +
				"	DOM		DATE,  " +
				"	PageSize	INT,  " +

				"	PRIMARY KEY (EntryId) " +
				")"
			);

			stmt.execute
			(
				"CREATE TABLE Link "+
				"(" +
				"	LinkId		INT		UNIQUE 	NOT NULL, " +
				"	Entry		INT		NOT NULL, " +
				"	Domain		VARCHAR(50) 	NOT NULL, " +
				"	Path		VARCHAR(100) 	NOT NULL, " +
				"	Page		VARCHAR(50), " +
				"	URL			VARCHAR(200)	NOT NULL," +
				"	LinksTo		INT,  " +

				"	PRIMARY KEY (LinkId) " +
				")"
			);

			con.close();


		}
		// --- connection could not be opened -> exit
		catch (Exception e)
		{
			System.out.println(e);
		}
	}

}