import java.util.*;
import java.io.*;
import java.sql.*;



/*------------------------------------------------------------- XMLparser class
|																			  |
|						Class to parse a GGLE xml file						  |
|																			  |
|																			  |
|																			  |
| 													by Felix Hageloh 30/09/04 |
\----------------------------------------------------------------------------*/

public class XMLparser
{
	Hashtable tuples = new Hashtable();	//stores all data for an entry

	private BufferedReader fileIn;		//the file reader
	public boolean inGGLE = false;		//true if inside ggle tag


	// ---------------------------------- constants for ggle tags
	public static final String GGLE = "GGLE";
	public static final String GGLEESC = "/GGLE";
    public static final String ENTR = "ENTRY";
	public static final String ENTRYESC = "/ENTRY";
    public static final String LINK = "LINK";
    public static final String DEL = "\""; // delimiter for attributes



	/*------------------------------------------------------------ get functions
	|																		   |
	| typical get functions													   |
	\-------------------------------------------------------------------------*/
	public String getURL()
	{
		return (String) tuples.get("URL");
	}

	public java.sql.Date getDOC()
	{
		long secs = Long.parseLong((String) tuples.get("DOC") );

		return (new java.sql.Date(secs) );
	}

	public java.sql.Date getDOM()
	{
		long secs = Long.parseLong((String) tuples.get("DOM") );

		return (new java.sql.Date(secs) );
	}

	public int getSIZE()
	{
		return (int) Integer.parseInt((String) tuples.get("SIZE") );
	}

	public Vector getLINKS()
	{
		return (Vector) tuples.get("LINKS");
	}

	/*--------------------------------------------- XMLparser( String FileName )
	|																		   |
	| Constructer: Creates a new xmlparser on the given file				   |																		   |
	\-------------------------------------------------------------------------*/
	public XMLparser( String strFname )
	{
		//try to open the file
		try
		{
			fileIn = new BufferedReader( new FileReader(strFname) );
		}
		//catch file not found exception
		catch(FileNotFoundException e)
		{
			System.out.println("Could not open " +strFname);
			fileIn = null;
		}
		//catch other exeptions -> close the file
		catch(Exception e)
		{
			try
			{
				fileIn.close();
			}
			catch(IOException e2)
			{
				System.out.println("couldn't close file");
			}
			fileIn = null;
		}
		System.out.println("Parsing file " + strFname + " ..." );
	}

	/*---------------------------------------------------------------- cleanUp()
	| 																		   |
	| closes the input stream												   |
	\-------------------------------------------------------------------------*/
	public void cleanUp ()
	{
		if( fileIn != null )
		{
			try
			{
				fileIn.close();
			}
			catch(IOException e2)
			{
				System.out.println("couldn't close file");
			}
		}


		fileIn = null;
	}


	/*--------------------------------------------------------------- goToGGLE()
	|																		   |
	| finds the next ggle tag. returns true if found, false otherwise		   |
	\-------------------------------------------------------------------------*/
	public boolean goToGGLE () throws IOException
	{
		int c = 0;
		String line = "";
		boolean found = false;

		inGGLE = false;

		if (fileIn != null )
		{
			while ( (c = fileIn.read()) != -1  && c != '<');

			if( c == '<' )
			{
				line = this.getLine('>');
				line = line.toUpperCase().trim();

				if( line.equals(GGLE) )
				{
					//System.out.println("found: " + line );
					found = true;
					inGGLE = true;
				}
			}
		}

		return found;
	}

	/*-------------------------------------------------------------- nextEntry()
	|																		   |
	| reads the next entry in the file and puts the data in the corresponding  |
	| class variables. 														   |
	| Returns: 	1 if entry was read sucessfully								   |
	| 			0 if entry was not read										   |
	| 		   -1 if eof													   |
	\-------------------------------------------------------------------------*/
	public int nextEntry() throws IOException
	{
		String line;
		String[] words;
		String[] tag;
		Vector links;

		int status = -1; // -1: EOF, 0: Fail, 1: Sucess

		// ----------- check if inside ggle tags, if not try to find <ggle> tag
		if( !inGGLE )
		{
			try
			{
				goToGGLE();
			}
			catch( IOException e)
			{
				System.out.println(e);
			}
		}


		// -------------------------------------------- if now inside ggle tags
		if( inGGLE )
		{
			words = new String[8];
			tag = new String[2];
			links = new Vector();


			// try to read the next entry
			line = getLine('>').trim();



			// --------------------------------------------- if we read someting
			if( line != "" )
			{
				line = line.substring(1);

				tag = line.split("\\s", 2);

				//System.out.println("Tag: " + tag[0] + " && " + tag[1] );

				// ------------------------------- if it is an entry tag procede
				if( tag[0].toUpperCase().equals(ENTR) )
				{
					words = tag[1].split(
						"\\s?\\=\\s?\\"+DEL+"\\s?|\\s?\\"+DEL+"\\s?");

					for( int i = 0; i < words.length - 1; i++ )
						tuples.put( words[i].toUpperCase(), words[i+1] );


					status = 2;

					// ----------------------------------------- read all links
					while( status > 1 )
					{
						line = getLine('>').trim().substring(1);

						// get the next tag
						tag = line.split("\\s", 2);

						//System.out.println("Tag: " + tag[0] + " && " + tag[1] );

						// -------------------------------- if link, read link
						if( tag[0].toUpperCase().equals(LINK) )
						{
							words = tag[1].split(
								"\\s?\\=\\s?\\"+DEL+"\\s?|\\s?\\"+DEL+"\\s?");

							// ---------------------- check if url is specified
							if( words[0].toUpperCase().equals("URL") )
								links.add(words[1]); //add link url
							// ------------------------------- else print error
							else
								System.out.println("ERROR: Link url expected.");

						}
						// ---------------------- if end of entry quit the loop
						else if( tag[0].toUpperCase().equals(ENTRYESC) )
						{
							status = 1;
						}
						//-------- else print error and quit the loop with error
						else
						{
							System.out.println("ERROR: Link  or /entry expected.");
							status = 0;
						}
					}// end while

					tuples.put("LINKS", links);
				}
				// ------------------------------------------------ end of ggle
				else if ( tag[0].toUpperCase().equals(GGLEESC) )
				{
					inGGLE = false;
					status = 0;
				}
				// ------------------------------- else print error and continue
				else
				{
					System.out.println("ERROR: Entry expected. Scipping");
					status = 0;
				}


			}

		}

		return status;
	}

	/*------------------------------------------------------- getLine( int del )
	|																		   |
	| reads one line from the input stream until 'del' is reached			   |
	\-------------------------------------------------------------------------*/
	public String getLine( int del ) throws IOException
	{
		int c;
		String line = "";

		while ( (c = fileIn.read()) != -1  && c != del)
		{
			line += (char)c;
		}

		return line;
	}


	/*------------------------------------------------------------------- main()
	|																		   |
	| main function															   |
	\-------------------------------------------------------------------------*/
	public static void main (String argv[]) throws IOException
	{
		int stat;
		XMLparser myP;

		myP = new XMLparser("test.xml");

		if( myP != null )
		{
			stat = myP.nextEntry();

			while( stat != -1 )
			{
				if ( stat == 1 )
				{
					System.out.println( "\n\nENTRY:");

					System.out.println( myP.getURL() );
					System.out.println( myP.getDOC() );
					System.out.println( myP.getDOM() );
					System.out.println( myP.getSIZE() );

					Vector v = myP.getLINKS();

					System.out.println( "\nLINKS: ");
					for (Enumeration e = v.elements() ; e.hasMoreElements() ;)
					{
         				System.out.print(e.nextElement() + ", ");
					}
				}

				stat = myP.nextEntry();
			}
		}

		myP.cleanUp();
	}
}
